/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with the program; if not, see <http://www.gnu.org/licenses/>
 *
 *
 * Copyright (C) 1999-2008 Novell, Inc. (www.novell.com)
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "e-ews-compat.h"

#include <string.h>
#include <glib/gi18n-lib.h>
#include <glib/gstdio.h>
#include <shell/e-shell.h>
#if EDS_CHECK_VERSION(2,33,0)
#include <mail/e-mail-session.h>
#else
#include <mail/mail-session.h>
#endif
#include <mail/e-mail-backend.h>

#include <camel/camel.h>
#include <libedataserver/e-account.h>
#include <libedataserver/e-account-list.h>
#include <libebook/e-book.h>
#include <camel-ews-utils.h>

#include "exchange-ews-account-listener.h"
#include "exchange-ews-account-setup.h"
#include "camel-ews-store-summary.h"
#include "ews-camel-compat.h"

#include "e-ews-config-utils.h"
#include "server/e-ews-folder.h"
#include "server/e-ews-gnome-3-8-compat.h"

#define d(x) x

#define EWS_BASE_URI "ews://"

gchar *	org_gnome_ews_get_password (EAccount *account, gboolean reask, const GError *error);

G_DEFINE_TYPE (ExchangeEWSAccountListener, exchange_ews_account_listener, G_TYPE_OBJECT)

static GObjectClass *parent_class = NULL;

struct _ExchangeEWSAccountListenerPrivate {
	GConfClient *gconf_client;
	EAccountList *account_list;
};

typedef struct _EwsAccountInfo EwsAccountInfo;

static void
ews_account_added (EAccountList *account_listener, EAccount *account);

struct _EwsAccountInfo {
	gchar *uid;
	gchar *name;
	gchar *source_url;
	gboolean enabled;
};

static	GList *ews_accounts = NULL;

static gboolean
is_ews_account (EAccount *account)
{
	return (account->source->url && (g_ascii_strncasecmp (account->source->url, EWS_URI_PREFIX, EWS_PREFIX_LENGTH) == 0));
}

static EwsAccountInfo*
lookup_account_info (const gchar *key)
{
	GList *list;

	g_return_val_if_fail (key != NULL, NULL);

	for (list = g_list_first (ews_accounts); list; list = g_list_next (list)) {
		EwsAccountInfo *info = (EwsAccountInfo *)(list->data);
		if (g_ascii_strcasecmp (info->uid, key) == 0)
			return info;
	}

	return NULL;
}

static EwsAccountInfo *
ews_account_info_from_eaccount (EAccount *account)
{
	EwsAccountInfo *info;

	info = g_new0 (EwsAccountInfo, 1);
	info->uid = g_strdup (account->uid);
	info->name = g_strdup (account->name);
	info->source_url = g_strdup (account->source->url);
	info->enabled = account->enabled;

	return info;
}

static void
ews_account_info_free (EwsAccountInfo *info)
{
	if (info) {
		g_free (info->uid);
		g_free (info->name);
		g_free (info->source_url);
		g_free (info);
	}
}

static gboolean
remove_group_in_source_list (const gchar *master_hosturl,
			     const gchar *master_username,
			     ESourceList *source_list)
{
	GSList *giter;

	g_return_val_if_fail (master_hosturl != NULL, FALSE);
	g_return_val_if_fail (master_username != NULL, FALSE);

	if (!source_list)
		return FALSE;

	for (giter = e_source_list_peek_groups (source_list); giter; giter = giter->next) {
		ESourceGroup *group = giter->data;

		if (group) {
			GSList *sources;

			sources = e_source_group_peek_sources (group);
			if (sources && sources->data &&
			    e_ews_folder_utils_source_is_for_account (sources->data, master_hosturl, master_username)) {
				return e_source_list_remove_group (source_list, group);
			}
		}
	}

	return FALSE;
}

static void
remove_groups_for_account (const gchar *setup_url)
{
	CamelURL *url;
	ESourceList *contacts = NULL, *events = NULL, *memos = NULL, *tasks = NULL;
	const gchar *master_hosturl, *master_username;
	gboolean changed = FALSE;

	g_return_if_fail (setup_url != NULL);
	g_return_if_fail (g_ascii_strncasecmp (setup_url, "ews://", 6) == 0);

	url = camel_url_new (setup_url, NULL);
	g_return_if_fail (url != NULL);

	e_ews_folder_utils_manage_source_lists (&contacts, &events, &memos, &tasks, FALSE);

	master_hosturl = camel_url_get_param (url, "hosturl");
	master_username = url->user;

	changed = remove_group_in_source_list (master_hosturl, master_username, contacts) || changed;
	changed = remove_group_in_source_list (master_hosturl, master_username, events) || changed;
	changed = remove_group_in_source_list (master_hosturl, master_username, memos) || changed;
	changed = remove_group_in_source_list (master_hosturl, master_username, tasks) || changed;

	if (changed)
		e_ews_folder_utils_manage_source_lists (NULL, NULL, NULL, NULL, TRUE);

	camel_url_free (url);
}

EVO2 (CamelSession *session;)

static void
ews_account_removed (EAccountList *account_listener, EAccount *account)
{
	EVO3(EShell *shell;)
	EVO3(EShellBackend *shell_backend;)
	EVO3(CamelSession *session;)
	EwsAccountInfo *info = NULL;
	CamelService *service;
	gchar *summary_file, *storage_path;

	if (!is_ews_account (account))
		return;

	info = lookup_account_info (account->uid);
	if (!info)
		return;

	remove_groups_for_account (e_account_get_string (account, E_ACCOUNT_SOURCE_URL));
	ews_accounts = g_list_remove (ews_accounts, info);

	EVO3(shell = e_shell_get_default ();)
	EVO3(shell_backend = e_shell_get_backend_by_name (shell, "mail");)
	EVO3 (session = (CamelSession *) e_mail_backend_get_session (E_MAIL_BACKEND (shell_backend));)
	service = camel_session_get_service_compat (session, account->source->url, CAMEL_PROVIDER_STORE);

	/* FIXME  Use this until CamelStore gets camel_store_remove_storage API which would be available eds 3.2 onwards */
	storage_path = camel_session_get_storage_path (session, service, NULL);
	summary_file = g_build_filename (storage_path, "folder-tree", NULL);
	g_unlink (summary_file);
	
	d(g_print ("Removed ews store summary: %s \n", summary_file);)
	
	g_free (storage_path);
	g_free (summary_file);
	ews_account_info_free (info);

#if ! EDS_CHECK_VERSION (3,1,0)
	g_object_unref (service);
#endif	
}

static void ews_update_folder_hierarchy (EAccount *account, const GError *with_error);

static void
ews_account_listener_folder_hierarchy_cb (GObject *cnc_object,
					  GAsyncResult *result,
					  gpointer user_data)
{
	EEwsConnection *connection;
	EAccount *account = user_data;
	GSList *folders_created = NULL;
	GSList *folders_updated = NULL;
	GSList *folders_deleted = NULL;
	gboolean includes_last_folder = FALSE;
	gchar *new_sync_state = NULL;
	gboolean success;
	GError *local_error = NULL;

	g_return_if_fail (account != NULL);
	g_return_if_fail (E_IS_EWS_CONNECTION (cnc_object));

	connection = E_EWS_CONNECTION (cnc_object);

	success = e_ews_connection_sync_folder_hierarchy_finish (
		connection,
		result,
		&new_sync_state,
		&includes_last_folder,
		&folders_created,
		&folders_updated,
		&folders_deleted,
		&local_error);

	if (success) {
		CamelURL *url;

		url = camel_url_new (e_account_get_string (account, E_ACCOUNT_SOURCE_URL), NULL);
		if (url) {
			gint ii;
			GSList *iter;
			GList *old_sources, *liter;
			GHashTable *old_sources_hash;
			GHashTableIter hiter;
			gpointer key, value;
			const gchar *master_hosturl, *master_username;

			master_hosturl = camel_url_get_param (url, "hosturl");
			master_username = url->user;

			old_sources_hash = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_object_unref);
			old_sources = e_ews_folder_utils_get_esources (master_hosturl, master_username, NULL, NULL);
			for (liter = old_sources; liter; liter = liter->next) {
				ESource *source = liter->data;
				const gchar *folder_id, *gal;

				if (!source)
					continue;

				folder_id = e_source_get_property (source, "folder-id");
				if (!folder_id || !*folder_id)
					continue;

				gal = e_source_get_property (source, "gal");
				if (gal && g_str_equal (gal, "1"))
					continue;

				g_hash_table_insert (old_sources_hash, g_strdup (folder_id), g_object_ref (source));
			}

			for (iter = folders_deleted; iter; iter = iter->next) {
				const gchar *folder_id = iter->data;

				if (folder_id && *folder_id) {
					g_hash_table_remove (old_sources_hash, folder_id);

					e_ews_folder_utils_remove_as_esource (
						master_hosturl, master_username,
						folder_id, NULL, NULL);
				}
			}

			for (ii = 0, iter = folders_created; iter; iter = iter->next, ii++) {
				EEwsFolder *folder = iter->data;
				EEwsFolderType folder_type;
				const EwsFolderId *folder_id;

				if (!folder)
					continue;

				folder_id = e_ews_folder_get_id (folder);
				if (!folder_id || !folder_id->id)
					continue;

				g_hash_table_remove (old_sources_hash, folder_id->id);

				folder_type = e_ews_folder_get_folder_type (folder);
				if (folder_type != E_EWS_FOLDER_TYPE_CALENDAR &&
				    folder_type != E_EWS_FOLDER_TYPE_CONTACTS &&
				    folder_type != E_EWS_FOLDER_TYPE_TASKS &&
				    folder_type != E_EWS_FOLDER_TYPE_MEMOS)
					continue;

				e_ews_folder_utils_add_as_esource (url, account->name,
					folder, 0, ii, NULL, NULL);
			}

			/* this might be always empty */
			for (iter = folders_updated; iter; iter = iter->next) {
				EEwsFolder *folder = iter->data;
				const EwsFolderId *folder_id;

				if (!folder)
					continue;

				folder_id = e_ews_folder_get_id (folder);
				if (!folder_id || !folder_id->id)
					continue;

				g_hash_table_remove (old_sources_hash, folder_id->id);
			}

			g_hash_table_iter_init (&hiter, old_sources_hash);
			while (g_hash_table_iter_next (&hiter, &key, &value)) {
				ESource *source = value;
				const gchar *tmp;

				if (!source)
					continue;

				/* keep foreign folders */
				tmp = e_source_get_property (source, "foreign");
				if (tmp && g_str_equal (tmp, "1"))
					continue;

				/* keep public folders */
				tmp = e_source_get_property (source, "public");
				if (tmp && g_str_equal (tmp, "1"))
					continue;

				tmp = e_source_get_property (source, "folder-id");
				if (!tmp || !*tmp)
					continue;

				/* drop all other untouched folders */
				e_ews_folder_utils_remove_as_esource (master_hosturl, master_username, tmp, NULL, NULL);
			}

			g_hash_table_destroy (old_sources_hash);
			g_list_free_full (old_sources, g_object_unref);
			camel_url_free (url);
		}

		/* List of EEwsFolder instances. */
		g_slist_free_full (folders_created, g_object_unref);
		/* List of folder ID strings. */
		g_slist_free_full (folders_deleted, g_free);
		/* List of EEwsFolder instances. */
		g_slist_free_full (folders_updated, g_object_unref);

	} else if (g_error_matches (local_error, EWS_CONNECTION_ERROR, EWS_CONNECTION_ERROR_INVALIDSYNCSTATEDATA)) {
		g_clear_error (&local_error);

		remove_groups_for_account (e_account_get_string (account, E_ACCOUNT_SOURCE_URL));

		e_ews_connection_sync_folder_hierarchy (
			connection,
			G_PRIORITY_DEFAULT,
			NULL,
			NULL,
			ews_account_listener_folder_hierarchy_cb,
			g_object_ref (account));
	} else if (g_error_matches (local_error, SOUP_HTTP_ERROR, SOUP_STATUS_UNAUTHORIZED) && account->enabled) {
		ews_update_folder_hierarchy (account, local_error);
	} else if (local_error) {
		g_warning ("%s: Failed with error: %s", G_STRFUNC, local_error->message);
	}

	g_clear_error (&local_error);
	g_object_unref (account);
}

static void
ews_update_folder_hierarchy (EAccount *account,
			     const GError *with_error)
{
	EwsAccountInfo *info;
	const gchar *timeout_str;
	gchar *password;
	CamelURL *url;
	EEwsConnection *connection;

	g_return_if_fail (account != NULL);

	if (!account->enabled ||
	    !e_ews_config_utils_is_online ())
		return;

	info = lookup_account_info (account->uid);
	if (!info)
		return;

	password = org_gnome_ews_get_password (account, with_error != NULL, with_error);
	if (!password)
		return;

	url = camel_url_new (e_account_get_string (account, E_ACCOUNT_SOURCE_URL), NULL);
	timeout_str = camel_url_get_param (url, "timeout");
	connection = e_ews_connection_new (camel_url_get_param (url, "hosturl"),
		url->user,
		password,
		camel_url_get_param (url, "email"),
		camel_url_get_param (url, "impersonate_user"),
		timeout_str ? atoi (timeout_str) : -1,
		TRUE,
		NULL, NULL, NULL);

	g_free (password);
	password = NULL;

	camel_url_free (url);

	if (connection) {
		e_ews_connection_sync_folder_hierarchy (
			connection,
			G_PRIORITY_DEFAULT,
			NULL,
			NULL,
			ews_account_listener_folder_hierarchy_cb,
			g_object_ref (account));
		g_object_unref (connection);
	} else
		g_warning ("%s: Failed to create EWS connection for account '%s'", G_STRFUNC, account->name);
}

static gboolean
run_account_sources_update_cb (gpointer user_data)
{
	EAccount *account = user_data;

	g_return_val_if_fail (account != NULL, FALSE);

	ews_update_folder_hierarchy (account, NULL);

	g_object_unref (account);

	return FALSE;
}

static gboolean
ews_is_str_equal (const gchar *str1, const gchar *str2)
{
	if (str1 && str2 && !strcmp (str1, str2))
		return TRUE;
	else if (!str1 && !str2)
		return TRUE;
	else
		return FALSE;
}

static void
remove_gal_esource (CamelURL *setup_url)
{
	GError *error = NULL;

	e_ews_folder_utils_remove_as_esource (camel_url_get_param (setup_url, "hosturl"),
		setup_url->user, "global-address-list", NULL, &error);

	if (error) {
		g_warning ("Unable to remove GAL cache: %s \n", error->message);
		g_clear_error (&error);
	}
}

/* add gal esource. If oal is not selected, gal will be just used for auto-completion */
static void
add_gal_esource (const gchar *account_name,
		 CamelURL *url)
{
	ESourceList *contacts = NULL;
	ESourceGroup *group;
	ESource *source;
	const gchar *email_id;
	const gchar *oal_sel, *tmp, *oal_name;
	gchar *oal_id = NULL;
	gchar *account_uri, *source_uri;
	GSList *giter;

	e_ews_folder_utils_manage_source_lists (&contacts, NULL, NULL, NULL, FALSE);
	email_id = camel_url_get_param (url, "email");
	oal_sel = camel_url_get_param (url, "oal_selected");

	/* if oal is not selected, gal just performs auto-completion and does not cache GAL */	
	if (oal_sel) {
		tmp = strrchr (oal_sel, ':');
		oal_name = tmp + 1;
		if (*oal_name == '\\')
			oal_name++;
			
		oal_id = g_strndup (oal_sel, (tmp - oal_sel));
	} else
		oal_name = _("Global Address list");

	account_uri = camel_url_to_string (url, CAMEL_URL_HIDE_PASSWORD | CAMEL_URL_HIDE_PARAMS);
	source_uri = g_strdup_printf ("%s;gal=1", account_uri + strlen (EWS_BASE_URI));
	source = e_source_new (oal_name, source_uri);
	
	/* set properties */
	e_source_set_property (source, "username", url->user);
	e_source_set_property (source, "auth-domain", "Exchange Web Services");
	e_source_set_property (source, "email", email_id);
	e_source_set_property (source, "gal", "1");
	e_source_set_property (source, "hosturl", camel_url_get_param (url, "hosturl"));
	e_source_set_property (source, "delete", "no");
	e_source_set_property (source, "folder-id", "global-address-list");

	/* If oal_id is present it means the GAL is marked for offline usage, we do not check for offline_sync property */
	if (oal_sel) {
		e_source_set_property (source, "oal_id", oal_id);
		e_source_set_property (source, "oab_url", camel_url_get_param (url, "oaburl"));
	}

	e_source_set_property (source, "auth", "plain/password");
	e_source_set_property (source, "completion", "true");

	/* add the source to group and sync */
	for (giter = e_source_list_peek_groups (contacts); giter; giter = giter->next) {
		group = giter->data;

		if (group && e_source_group_peek_base_uri (group) &&
		    g_ascii_strcasecmp (e_source_group_peek_base_uri (group), EWS_BASE_URI) == 0 &&
		    e_source_group_peek_name (group) &&
		    g_str_equal (e_source_group_peek_name (group), account_name)) {
			g_object_ref (group);
			break;
		}

		group = NULL;
	}

	if (!group) {
		group = e_source_group_new (account_name, EWS_BASE_URI);
		e_source_group_set_property (group, "create_source", "no");

		e_source_list_add_group (contacts, group, -1);
	}

	e_source_group_add_source (group, source, -1);

	e_source_list_sync (contacts, NULL);

	g_object_unref (source);
	g_object_unref (group);
	g_free (oal_id);
	g_free (account_uri);
	g_free (source_uri);

	return;
}

static void
ews_account_changed (EAccountList *account_listener, EAccount *account)
{
	gboolean ews_account = FALSE;
	EwsAccountInfo *existing_account_info = NULL;

	ews_account = is_ews_account (account);

	if (ews_account)
		existing_account_info = lookup_account_info (account->uid);

	if (existing_account_info == NULL && ews_account && account->enabled) {
		ews_account_added (account_listener, account);
	} else if (existing_account_info != NULL && !ews_account)
		ews_account_removed (account_listener, account);
	else if (existing_account_info != NULL && ews_account) {
		if (!account->enabled)
			ews_account_removed (account_listener, account);
		else {
			CamelURL *old_url, *new_url;
			const gchar *o_oal_sel, *n_oal_sel;
			
			/* TODO update props like refresh timeout */
			old_url = camel_url_new (existing_account_info->source_url, NULL);
			new_url = camel_url_new (account->source->url, NULL);

			o_oal_sel = camel_url_get_param (old_url, "oal_selected");
			n_oal_sel = camel_url_get_param (new_url, "oal_selected");

			if (!ews_is_str_equal (o_oal_sel, n_oal_sel)) {
				/* remove gal esource and cache associated with it */
				remove_gal_esource (old_url);

				/* add gal esource */
				add_gal_esource (account->name, new_url);
			}
			
			g_free (existing_account_info->name);
			g_free (existing_account_info->source_url);
			existing_account_info->name = g_strdup (account->name);
			existing_account_info->source_url = g_strdup (account->source->url);

			camel_url_free (old_url);
			camel_url_free (new_url);
		}
	}
}

static void
ews_account_added (EAccountList *account_listener, EAccount *account)
{
	gboolean ews_account = FALSE;

	ews_account = is_ews_account (account);

	if (ews_account) {
		CamelURL *url;

		EwsAccountInfo *info = ews_account_info_from_eaccount (account);
		ews_accounts = g_list_append (ews_accounts, info);
		url = camel_url_new (account->source->url, NULL);
		
		/* add gal esource */
		add_gal_esource (account->name, url);
		ews_update_folder_hierarchy (account, NULL);

		camel_url_free (url);
	}
}

static void
exchange_ews_account_listener_construct (ExchangeEWSAccountListener *config_listener)
{
	EIterator *iter;

	d(g_print ("\n Construct the listener"));

	config_listener->priv->account_list = e_account_list_new (config_listener->priv->gconf_client);

	for (iter = e_list_get_iterator (E_LIST(config_listener->priv->account_list)); e_iterator_is_valid (iter); e_iterator_next (iter)) {
		EAccount *account = E_ACCOUNT (e_iterator_get (iter));
		if (is_ews_account (account) && account->enabled) {
			EwsAccountInfo *info;

			info = ews_account_info_from_eaccount (account);
			ews_accounts = g_list_append (ews_accounts, info);

			g_idle_add (run_account_sources_update_cb, g_object_ref (account));
		}
	}

	g_signal_connect (config_listener->priv->account_list, "account_added", G_CALLBACK (ews_account_added), NULL);
	g_signal_connect (config_listener->priv->account_list, "account_changed", G_CALLBACK (ews_account_changed), NULL);
	g_signal_connect (config_listener->priv->account_list, "account_removed", G_CALLBACK (ews_account_removed), NULL);
}

static void
dispose (GObject *object)
{
	ExchangeEWSAccountListener *config_listener = EXCHANGE_EWS_ACCOUNT_LISTENER (object);

	g_object_unref (config_listener->priv->gconf_client);
	g_object_unref (config_listener->priv->account_list);

	G_OBJECT_CLASS (parent_class)->dispose (object);
}

static void
finalize (GObject *object)
{
	ExchangeEWSAccountListener *config_listener = EXCHANGE_EWS_ACCOUNT_LISTENER (object);
	GList *list;

	if (config_listener->priv) {
		g_free (config_listener->priv);
	}

	for (list = g_list_first (ews_accounts); list; list = g_list_next (list)) {
		EwsAccountInfo *info = (EwsAccountInfo *)(list->data);
		ews_account_info_free (info);
	}

	g_list_free (ews_accounts);
	ews_accounts = NULL;

	G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
exchange_ews_account_listener_class_init (ExchangeEWSAccountListenerClass *class)
{
	GObjectClass *object_class;

	parent_class = g_type_class_ref (G_TYPE_OBJECT);
	object_class = G_OBJECT_CLASS (class);

	/* virtual method override */
	object_class->dispose = dispose;
	object_class->finalize = finalize;
}

static void
exchange_ews_account_listener_init (ExchangeEWSAccountListener *config_listener)
{
	config_listener->priv = g_new0 (ExchangeEWSAccountListenerPrivate, 1);
}

ExchangeEWSAccountListener *
exchange_ews_account_listener_new (void)
{
	ExchangeEWSAccountListener *config_listener;

	config_listener = g_object_new (EXCHANGE_EWS_ACCOUNT_LISTENER_TYPE, NULL);
	config_listener->priv->gconf_client = gconf_client_get_default();

	exchange_ews_account_listener_construct (config_listener);

	return config_listener;
}
