/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * Authors :
 *
 * Copyright (C) 1999-2008 Novell, Inc. (www.novell.com)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU Lesser General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301
 * USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string.h>
#include <glib/gi18n-lib.h>

#include <libedataserver/e-source-group.h>
#include <libedataserver/e-source.h>
#include <libedataserver/e-source-list.h>

#include "e-ews-folder.h"
#include "e-ews-message.h"
#include "e-ews-enumtypes.h"
#include "ews-errors.h"

#include "e-ews-gnome-3-8-compat.h"

G_DEFINE_TYPE (EEwsFolder, e_ews_folder, G_TYPE_OBJECT)

struct _EEwsFolderPrivate {
	GError *error;
	gchar *name;
	EwsFolderId *fid;
	EwsFolderId *parent_fid;
	EEwsFolderType folder_type;
	guint32 unread;
	guint32 total;
	guint32 child_count;
	gboolean foreign;
};

static GObjectClass *parent_class = NULL;

static void
e_ews_folder_dispose (GObject *object)
{
	EEwsFolder *folder = (EEwsFolder *) object;

	g_return_if_fail (E_IS_EWS_FOLDER (folder));

	if (parent_class->dispose)
		(* parent_class->dispose) (object);
}

static void
e_ews_folder_finalize (GObject *object)
{
	EEwsFolder *folder = (EEwsFolder *) object;
	EEwsFolderPrivate *priv;

	g_return_if_fail (E_IS_EWS_FOLDER (folder));

	priv = folder->priv;

	g_clear_error (&priv->error);

	if (priv->name) {
		g_free (priv->name);
		priv->name = NULL;
	}

	if (priv->fid) {
		g_free (priv->fid->id);
		g_free (priv->fid->change_key);
		g_free (priv->fid);
		priv->fid = NULL;
	}

	if (priv->parent_fid) {
		g_free (priv->parent_fid->id);
		g_free (priv->parent_fid->change_key);
		g_free (priv->parent_fid);
		priv->parent_fid = NULL;
	}

	g_free (priv);
	folder->priv = NULL;

	if (parent_class->finalize)
		(* parent_class->finalize) (object);
}

static void
e_ews_folder_class_init (EEwsFolderClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

	parent_class = g_type_class_peek_parent (klass);

	object_class->dispose = e_ews_folder_dispose;
	object_class->finalize = e_ews_folder_finalize;
}

static void
e_ews_folder_init (EEwsFolder *folder)
{
	EEwsFolderPrivate *priv;

	/* allocate internal structure */
	priv = g_new0 (EEwsFolderPrivate, 1);
	folder->priv = priv;

	priv->error = NULL;
	priv->folder_type = E_EWS_FOLDER_TYPE_UNKNOWN;
	priv->foreign = FALSE;
}

static gboolean
e_ews_folder_set_from_soap_parameter (EEwsFolder *folder,
                                      ESoapParameter *param)
{
	EEwsFolderPrivate *priv = folder->priv;
	ESoapParameter *subparam, *node;

	g_return_val_if_fail (param != NULL, FALSE);

	if (g_strcmp0 (e_soap_parameter_get_name (param), "Folder") == 0) {
		node = param;
		priv->folder_type = E_EWS_FOLDER_TYPE_MAILBOX;
	} else if (g_strcmp0 (e_soap_parameter_get_name (param), "CalendarFolder") == 0) {
		node = param;
		priv->folder_type = E_EWS_FOLDER_TYPE_CALENDAR;
	} else if (g_strcmp0 (e_soap_parameter_get_name (param), "ContactsFolder") == 0) {
		node = param;
		priv->folder_type = E_EWS_FOLDER_TYPE_CONTACTS;
	} else if (g_strcmp0 (e_soap_parameter_get_name (param), "TasksFolder") == 0) {
		node = param;
		priv->folder_type = E_EWS_FOLDER_TYPE_TASKS;
	} else if ((node = e_soap_parameter_get_first_child_by_name (param, "Folder")))
		priv->folder_type = E_EWS_FOLDER_TYPE_MAILBOX;
	else if ((node = e_soap_parameter_get_first_child_by_name (param, "CalendarFolder")))
		priv->folder_type = E_EWS_FOLDER_TYPE_CALENDAR;
	else if ((node = e_soap_parameter_get_first_child_by_name (param, "ContactsFolder")))
		priv->folder_type = E_EWS_FOLDER_TYPE_CONTACTS;
	else if ((node = e_soap_parameter_get_first_child_by_name (param, "TasksFolder")))
		priv->folder_type = E_EWS_FOLDER_TYPE_TASKS;
	else {
		g_warning ("Unable to find the Folder node \n");
		return FALSE;
	}

	if (priv->folder_type == E_EWS_FOLDER_TYPE_MAILBOX) {
		subparam = e_soap_parameter_get_first_child_by_name (node, "FolderClass");
		if (subparam) {
			gchar *folder_class = e_soap_parameter_get_string_value (subparam);

			priv->folder_type = E_EWS_FOLDER_TYPE_UNKNOWN;

			if (g_strcmp0 (folder_class, "IPF.Note") == 0 || (folder_class && g_str_has_prefix (folder_class, "IPF.Note."))) {
				priv->folder_type = E_EWS_FOLDER_TYPE_MAILBOX;
			} else if (g_strcmp0 (folder_class, "IPF.Contact") == 0) {
				priv->folder_type = E_EWS_FOLDER_TYPE_CONTACTS;
			} else if (g_strcmp0 (folder_class, "IPF.Appointment") == 0) {
				priv->folder_type = E_EWS_FOLDER_TYPE_CALENDAR;
			} else if (g_strcmp0 (folder_class, "IPF.Task") == 0) {
				priv->folder_type = E_EWS_FOLDER_TYPE_TASKS;
			} else if (g_strcmp0 (folder_class, "IPF.StickyNote") == 0) {
				priv->folder_type = E_EWS_FOLDER_TYPE_MEMOS;
			}

			g_free (folder_class);
		}
	}

	subparam = e_soap_parameter_get_first_child_by_name (node, "FolderId");
	if (subparam) {
		priv->fid = g_new0 (EwsFolderId, 1);
		priv->fid->id = e_soap_parameter_get_property (subparam, "Id");
		priv->fid->change_key = e_soap_parameter_get_property (subparam, "ChangeKey");
	}

	subparam = e_soap_parameter_get_first_child_by_name (node, "ParentFolderId");
	if (subparam) {
		priv->parent_fid = g_new0 (EwsFolderId, 1);
		priv->parent_fid->id = e_soap_parameter_get_property (subparam, "Id");
		priv->parent_fid->change_key = e_soap_parameter_get_property (subparam, "ChangeKey");
	}

	subparam = e_soap_parameter_get_first_child_by_name (node, "DisplayName");
	if (subparam)
		priv->name = e_soap_parameter_get_string_value (subparam);

	subparam = e_soap_parameter_get_first_child_by_name (node, "UnreadCount");
	if (subparam)
		priv->unread = e_soap_parameter_get_int_value (subparam);

	subparam = e_soap_parameter_get_first_child_by_name (node, "TotalCount");
	if (subparam)
		priv->total = e_soap_parameter_get_int_value (subparam);

	subparam = e_soap_parameter_get_first_child_by_name (node, "ChildFolderCount");
	if (subparam)
		priv->child_count = e_soap_parameter_get_int_value (subparam);

	return TRUE;
}

const gchar *
e_ews_folder_type_to_nick (EEwsFolderType folder_type)
{
	GEnumClass *enum_class;
	GEnumValue *enum_value;
	const gchar *folder_type_nick;

	enum_class = g_type_class_ref (E_TYPE_EWS_FOLDER_TYPE);
	enum_value = g_enum_get_value (enum_class, folder_type);

	if (enum_value == NULL) {
		folder_type = E_EWS_FOLDER_TYPE_UNKNOWN;
		enum_value = g_enum_get_value (enum_class, folder_type);
	}

	g_return_val_if_fail (enum_value != NULL, NULL);

	folder_type_nick = g_intern_string (enum_value->value_nick);

	g_type_class_unref (enum_class);

	return folder_type_nick;
}

EEwsFolderType
e_ews_folder_type_from_nick (const gchar *folder_type_nick)
{
	GEnumClass *enum_class;
	GEnumValue *enum_value;
	EEwsFolderType folder_type;

	g_return_val_if_fail (
		folder_type_nick != NULL,
		E_EWS_FOLDER_TYPE_UNKNOWN);

	enum_class = g_type_class_ref (E_TYPE_EWS_FOLDER_TYPE);
	enum_value = g_enum_get_value_by_nick (enum_class, folder_type_nick);

	if (enum_value != NULL)
		folder_type = enum_value->value;
	else
		folder_type = E_EWS_FOLDER_TYPE_UNKNOWN;

	g_type_class_unref (enum_class);

	return folder_type;
}

EEwsFolder *
e_ews_folder_new_from_soap_parameter (ESoapParameter *param)
{
	EEwsFolder *folder;

	g_return_val_if_fail (param != NULL, NULL);

	folder = g_object_new (E_TYPE_EWS_FOLDER, NULL);
	if (!e_ews_folder_set_from_soap_parameter (folder, param)) {
		g_object_unref (folder);
		return NULL;
	}

	return folder;
}

EEwsFolder *
e_ews_folder_new_from_error (const GError *error)
{
	EEwsFolder *folder;

	g_return_val_if_fail (error != NULL, NULL);

	folder = g_object_new (E_TYPE_EWS_FOLDER, NULL);
	folder->priv->error = g_error_copy (error);

	return folder;
}

gboolean
e_ews_folder_is_error (EEwsFolder *folder)
{
	g_return_val_if_fail (E_IS_EWS_FOLDER (folder), TRUE);

	return folder->priv->error != NULL;
}

const GError *
e_ews_folder_get_error (EEwsFolder *folder)
{
	g_return_val_if_fail (E_IS_EWS_FOLDER (folder), NULL);

	return folder->priv->error;
}

EwsFolderId *
e_ews_folder_id_new (const gchar *id,
                     const gchar *change_key,
                     gboolean is_distinguished_id)
{
	EwsFolderId *fid;

	fid = g_new0 (EwsFolderId, 1);
	fid->id = g_strdup (id);
	fid->change_key = g_strdup (change_key);
	fid->is_distinguished_id = is_distinguished_id;

	return fid;
}

void
e_ews_folder_id_free (EwsFolderId *fid)
{
	if (fid) {
		g_free (fid->id);
		g_free (fid->change_key);
		g_free (fid);
	}
}

const gchar *
e_ews_folder_get_name (EEwsFolder *folder)
{
	g_return_val_if_fail (E_IS_EWS_FOLDER (folder), NULL);

	return (const gchar *) folder->priv->name;
}

void
e_ews_folder_set_name (EEwsFolder *folder,
                       const gchar *new_name)
{
	EEwsFolderPrivate *priv;

	g_return_if_fail (E_IS_EWS_FOLDER (folder));
	g_return_if_fail (new_name != NULL);

	priv = folder->priv;

	if (priv->name)
		g_free (priv->name);
	priv->name = g_strdup (new_name);
}

const EwsFolderId *
e_ews_folder_get_id (EEwsFolder *folder)
{
	g_return_val_if_fail (E_IS_EWS_FOLDER (folder), NULL);

	return (const EwsFolderId *) folder->priv->fid;
}

const EwsFolderId *
e_ews_folder_get_parent_id (EEwsFolder *folder)
{
	g_return_val_if_fail (E_IS_EWS_FOLDER (folder), NULL);

	return (const EwsFolderId *) folder->priv->parent_fid;
}

void
e_ews_folder_set_parent_id (EEwsFolder *folder,
                            EwsFolderId *parent_fid)
{
	EEwsFolderPrivate *priv;

	g_return_if_fail (E_IS_EWS_FOLDER (folder));
	g_return_if_fail (parent_fid != NULL);

	priv = folder->priv;

	if (priv->parent_fid) {
		g_free (priv->parent_fid->id);
		g_free (priv->parent_fid->change_key);
		g_free (priv->parent_fid);
	}

	priv->parent_fid = parent_fid;
}

EEwsFolderType
e_ews_folder_get_folder_type (EEwsFolder *folder)
{
	g_return_val_if_fail (E_IS_EWS_FOLDER (folder), -1);

	return folder->priv->folder_type;
}

void
e_ews_folder_set_folder_type (EEwsFolder *folder,
                              EEwsFolderType folder_type)
{
	g_return_if_fail (E_IS_EWS_FOLDER (folder));

	folder->priv->folder_type = folder_type;
}

guint32
e_ews_folder_get_total_count (EEwsFolder *folder)
{
	g_return_val_if_fail (E_IS_EWS_FOLDER (folder), -1);

	return folder->priv->total;
}

guint32
e_ews_folder_get_unread_count (EEwsFolder *folder)
{
	g_return_val_if_fail (E_IS_EWS_FOLDER (folder), -1);

	return folder->priv->unread;
}

guint32
e_ews_folder_get_child_count (EEwsFolder *folder)
{
	g_return_val_if_fail (E_IS_EWS_FOLDER (folder), -1);

	return folder->priv->child_count;
}

gboolean
e_ews_folder_get_foreign (EEwsFolder *folder)
{
	g_return_val_if_fail (E_IS_EWS_FOLDER (folder), FALSE);

	return folder->priv->foreign;
}

void
e_ews_folder_set_foreign (EEwsFolder *folder,
                          gboolean is_foreign)
{
	g_return_if_fail (E_IS_EWS_FOLDER (folder));

	folder->priv->foreign = is_foreign;
}

gchar *
e_ews_folder_utils_pick_color_spec (gint move_by,
                                    gboolean around_middle)
{
	static gint color_mover = 0;
	static gint color_indexer = -1;
	const guint32 colors[] = {
		0x1464ae, /* dark blue */
		0x14ae64, /* dark green */
		0xae1464, /* dark red */
		0
	};
	guint32 color;

	if (move_by <= 0)
		move_by = 1;

	while (move_by > 0) {
		move_by--;

		color_indexer++;
		if (colors[color_indexer] == 0) {
			color_mover += 1;
			color_indexer = 0;
		}
	}

	color = colors[color_indexer];
	color = (color & ~(0xFF << (color_indexer * 8))) |
		(((((color >> (color_indexer * 8)) & 0xFF) + (0x33 * color_mover)) % 0xFF) << (color_indexer * 8));

	if (around_middle) {
		gint rr, gg, bb, diff;

		rr = (0xFF0000 & color) >> 16;
		gg = (0x00FF00 & color) >>  8;
		bb = (0x0000FF & color);

		diff = 0x80 - rr;
		if (diff < 0x80 - gg)
			diff = 0x80 - gg;
		if (diff < 0x80 - bb)
			diff = 0x80 - bb;

		rr = rr + diff < 0 ? 0 : rr + diff > 0xCC ? 0xCC : rr + diff;
		gg = gg + diff < 0 ? 0 : gg + diff > 0xCC ? 0xCC : gg + diff;
		bb = bb + diff < 0 ? 0 : bb + diff > 0xCC ? 0xCC : bb + diff;

		color = (rr << 16) + (gg << 8) + bb;
	}

	return g_strdup_printf ("#%06x", color);
}

#define EWS_BASE_URI "ews://"
/* do not unref set ESourceList-s */
void
e_ews_folder_utils_manage_source_lists (ESourceList **pcontacts,
					ESourceList **pevents,
					ESourceList **pmemos,
					ESourceList **ptasks,
					gboolean save_them)
{
	static GStaticRecMutex lock = G_STATIC_REC_MUTEX_INIT;
	static ESourceList *contacts = NULL, *events = NULL, *memos = NULL, *tasks = NULL;

	g_static_rec_mutex_lock (&lock);
	if (!contacts)
		contacts = e_source_list_new_for_gconf_default ("/apps/evolution/addressbook/sources");
	if (!events)
		events = e_source_list_new_for_gconf_default ("/apps/evolution/calendar/sources");
	if (!memos)
		memos = e_source_list_new_for_gconf_default ("/apps/evolution/memos/sources");
	if (!tasks)
		tasks = e_source_list_new_for_gconf_default ("/apps/evolution/tasks/sources");

	if (pcontacts)
		*pcontacts = contacts;
	if (pevents)
		*pevents = events;
	if (pmemos)
		*pmemos = memos;
	if (ptasks)
		*ptasks = tasks;

	if (save_them) {
		if (contacts)
			e_source_list_sync (contacts, NULL);
		if (events)
			e_source_list_sync (events, NULL);
		if (memos)
			e_source_list_sync (memos, NULL);
		if (tasks)
			e_source_list_sync (tasks, NULL);
	}

	g_static_rec_mutex_unlock (&lock);
}

static gboolean
e_ews_folder_save_source_lists_cb (gpointer user_data)
{
	guint *pscheduled = user_data;

	g_return_val_if_fail (pscheduled != NULL, FALSE);

	*pscheduled = 0;

	e_ews_folder_utils_manage_source_lists (NULL, NULL, NULL, NULL, TRUE);

	return FALSE;
}

static void
e_ews_folder_schedule_save_source_lists (void)
{
	static guint scheduled = 0;

	if (scheduled)
		g_source_remove (scheduled);

	scheduled = g_timeout_add_seconds (5, e_ews_folder_save_source_lists_cb, &scheduled);
}

static gboolean
e_ews_folder_source_uid_select_cb (gpointer user_data)
{
	gchar *source_uid = user_data;
	const gchar *selected_key = NULL;
	ESourceList *events = NULL, *memos = NULL, *tasks = NULL;

	g_return_val_if_fail (source_uid != NULL, FALSE);

	e_ews_folder_utils_manage_source_lists (NULL, &events, &memos, &tasks, FALSE);
	if (events && e_source_list_peek_source_by_uid (events, source_uid))
		selected_key = "/apps/evolution/calendar/display/selected_calendars";
	else if (memos && e_source_list_peek_source_by_uid (memos, source_uid))
		selected_key = "/apps/evolution/calendar/memos/selected_memos";
	else if (tasks && e_source_list_peek_source_by_uid (tasks, source_uid))
		selected_key = "/apps/evolution/calendar/tasks/selected_tasks";

	if (selected_key) {
		GConfClient *gconf_client = gconf_client_get_default ();
		GSList *ids;

		ids = gconf_client_get_list (gconf_client, selected_key , GCONF_VALUE_STRING, NULL);
		ids = g_slist_append (ids, g_strdup (source_uid));
		gconf_client_set_list (gconf_client,  selected_key, GCONF_VALUE_STRING, ids, NULL);

		g_slist_free_full (ids, g_free);
		g_object_unref (gconf_client);
	}

	g_free (source_uid);

	return FALSE;
}

static void
e_ews_folder_schedule_source_uid_select (const gchar *source_uid)
{
	g_idle_add (e_ews_folder_source_uid_select_cb, g_strdup (source_uid));
}

gboolean
e_ews_folder_utils_populate_esource (ESource *source,
                                     const GList *sources,
                                     CamelURL *setup_url,
                                     EEwsFolder *folder,
				     EEwsESourceFlags flags,
                                     gint color_seed,
                                     GCancellable *cancellable,
                                     GError **perror)
{
	EEwsFolderType folder_type;
	const EwsFolderId *folder_id = e_ews_folder_get_id (folder);

	g_return_val_if_fail (setup_url != NULL, FALSE);
	g_return_val_if_fail (folder_id != NULL, FALSE);

	folder_type = e_ews_folder_get_folder_type (folder);

	switch (folder_type) {
		case E_EWS_FOLDER_TYPE_CALENDAR:
		case E_EWS_FOLDER_TYPE_MEMOS:
		case E_EWS_FOLDER_TYPE_TASKS:
		case E_EWS_FOLDER_TYPE_CONTACTS:
			break;
		default:
			g_propagate_error (
				perror, g_error_new_literal (EWS_CONNECTION_ERROR,
				EWS_CONNECTION_ERROR_NORESPONSE, _("Cannot add folder, unsupported folder type")));
			return FALSE;
	}

	e_source_set_name (source, e_ews_folder_get_name (folder));

	e_source_set_property (source, "delete", "no");
	e_source_set_property (source, "username", setup_url->user);
	e_source_set_property (source, "auth-domain", "Exchange Web Services");
	e_source_set_property (source, "email", camel_url_get_param (setup_url, "email"));
	e_source_set_property (source, "hosturl", camel_url_get_param (setup_url, "hosturl"));
	e_source_set_property (source, "impersonate_user", camel_url_get_param (setup_url, "impersonate_user"));
	e_source_set_property (source, "timeout", camel_url_get_param (setup_url, "timeout"));

	e_source_set_property (source, "folder-id", folder_id->id);
	e_source_set_property (source, "change-key", NULL);
	e_source_set_property (source, "foreign", e_ews_folder_get_foreign (folder) ? "1" : NULL);
	e_source_set_property (source, "foreign_subfolders", (flags & E_EWS_ESOURCE_FLAG_INCLUDE_SUBFOLDERS) != 0 ? "1" : NULL);
	e_source_set_property (source, "public", (flags & E_EWS_ESOURCE_FLAG_PUBLIC_FOLDER) != 0 ? "1" : NULL);
	e_source_set_property (source, "offline_sync", (flags & E_EWS_ESOURCE_FLAG_OFFLINE_SYNC) != 0 ? "1" : NULL);

	if (folder_type == E_EWS_FOLDER_TYPE_CONTACTS) {
		e_source_set_property (source, "auth", "plain/password");
		e_source_set_property (source, "completion", "true");
	} else {
		gchar *color_str;
		gchar *today;
		GTimeVal today_tv;
		GDate dt;

		color_str = e_ews_folder_utils_pick_color_spec (
			1 + g_list_length ((GList *) sources),
			folder_type != E_EWS_FOLDER_TYPE_CALENDAR);
		e_source_set_color_spec (source, color_str);
		g_free (color_str);

		g_date_clear (&dt, 1);
		g_get_current_time (&today_tv);
		g_date_set_time_val (&dt, &today_tv);

		/* midnight UTC */
		today = g_strdup_printf ("%04d-%02d-%02dT00:00:00Z", g_date_get_year (&dt), g_date_get_month (&dt), g_date_get_day (&dt));

		e_source_set_property (source, "auth", "1");
		e_source_set_property (source, "alarm", "true");
		e_source_set_property (source, "last-notified", today);

		g_free (today);
	}

	return TRUE;
}

gboolean
e_ews_folder_utils_add_as_esource (CamelURL *setup_url,
				   const gchar *account_name,
                                   EEwsFolder *folder,
				   EEwsESourceFlags flags,
                                   gint color_seed,
                                   GCancellable *cancellable,
                                   GError **perror)
{
	GList *sources;
	ESource *source, *old_source;
	const EwsFolderId *fid;
	gboolean res = FALSE;

	g_return_val_if_fail (setup_url != NULL, FALSE);
	g_return_val_if_fail (account_name != NULL, FALSE);

	fid = e_ews_folder_get_id (folder);
	sources = e_ews_folder_utils_get_esources (camel_url_get_param (setup_url, "hosturl"), setup_url->user, cancellable, perror);

	old_source = e_ews_folder_utils_get_source_for_folder (sources, camel_url_get_param (setup_url, "hosturl"), setup_url->user, fid->id);
	if (old_source) {
		res = FALSE;

		g_propagate_error (
			perror,
			g_error_new (EWS_CONNECTION_ERROR, EWS_CONNECTION_ERROR_FOLDEREXISTS,
			_("Cannot add folder, folder already exists as '%s'"), e_source_peek_name (old_source)));
	} else {
		ESourceList *source_list = NULL;
		ESourceGroup *group = NULL;
		gchar *relative_uri;
		gchar *source_uri;
		GSList *giter;

		switch (e_ews_folder_get_folder_type (folder)) {
			case E_EWS_FOLDER_TYPE_CALENDAR:
				e_ews_folder_utils_manage_source_lists (NULL, &source_list, NULL, NULL, FALSE);
				break;
			case E_EWS_FOLDER_TYPE_MEMOS:
				e_ews_folder_utils_manage_source_lists (NULL, NULL, &source_list, NULL, FALSE);
				break;
			case E_EWS_FOLDER_TYPE_TASKS:
				e_ews_folder_utils_manage_source_lists (NULL, NULL, NULL, &source_list, FALSE);
				break;
			case E_EWS_FOLDER_TYPE_CONTACTS:
				e_ews_folder_utils_manage_source_lists (&source_list, NULL, NULL, NULL, FALSE);
				break;
			default:
				source_list = NULL;
				break;
		}

		if (!source_list) {
			g_propagate_error (
				perror, g_error_new_literal (EWS_CONNECTION_ERROR,
				EWS_CONNECTION_ERROR_NORESPONSE, _("Cannot add folder, unsupported folder type")));
			return FALSE;
		}

		for (giter = e_source_list_peek_groups (source_list); giter; giter = giter->next) {
			group = giter->data;

			if (group && e_source_group_peek_base_uri (group) &&
			    g_ascii_strcasecmp (e_source_group_peek_base_uri (group), EWS_BASE_URI) == 0 &&
			    e_source_group_peek_name (group) &&
			    g_str_equal (e_source_group_peek_name (group), account_name)) {
				g_object_ref (group);
				break;
			}

			group = NULL;
		}

		if (!group) {
			group = e_source_group_new (account_name, EWS_BASE_URI);
			e_source_group_set_property (group, "create_source", "no");

			e_source_list_add_group (source_list, group, -1);
		}

		relative_uri = camel_url_to_string (setup_url, CAMEL_URL_HIDE_ALL);
		g_return_val_if_fail (relative_uri != NULL, FALSE);
		g_return_val_if_fail (g_ascii_strncasecmp (relative_uri, EWS_BASE_URI, strlen (EWS_BASE_URI)) == 0, FALSE);

		source_uri = g_strdup_printf ("%s;folderid=%s", relative_uri, fid->id);
		g_free (relative_uri);
		relative_uri = source_uri;

		source = e_source_new (e_ews_folder_get_name (folder), relative_uri + strlen (EWS_BASE_URI));
		g_free (relative_uri);

		if (e_ews_folder_utils_populate_esource (
			source,
			sources,
			setup_url,
			folder,
			flags,
			color_seed,
			cancellable,
			perror)) {
			e_source_group_add_source (group, source, -1);
			e_ews_folder_schedule_save_source_lists ();
			e_ews_folder_schedule_source_uid_select (e_source_peek_uid (source));
			res = TRUE;
		}

		g_object_unref (source);
		g_object_unref (group);
	}

	g_list_free_full (sources, g_object_unref);

	return res;
}

gboolean
e_ews_folder_utils_remove_as_esource (const gchar *master_hosturl,
                                      const gchar *master_username,
                                      const gchar *folder_id,
                                      GCancellable *cancellable,
                                      GError **perror)
{
	ESource *source;
	GList *sources;
	gboolean res = TRUE;

	sources = e_ews_folder_utils_get_esources (master_hosturl, master_username, cancellable, perror);
	source = e_ews_folder_utils_get_source_for_folder (sources, master_hosturl, master_username, folder_id);

	if (source) {
		ESourceGroup *group = e_source_peek_group (source);

		if (group && e_source_group_remove_source (group, source)) {
			if (!e_source_group_peek_sources (group)) {
				ESourceList *contacts = NULL, *events = NULL, *memos = NULL, *tasks = NULL;

				e_ews_folder_utils_manage_source_lists (&contacts, &events, &memos, &tasks, FALSE);
				if (!e_source_list_remove_group (contacts, group))
					if (!e_source_list_remove_group (events, group))
						if (!e_source_list_remove_group (memos, group))
							e_source_list_remove_group (tasks, group);
			}

			e_ews_folder_schedule_save_source_lists ();
		}
	}

	g_list_free_full (sources, g_object_unref);

	return res;
}

gboolean
e_ews_folder_utils_source_is_for_account (ESource *source,
					  const gchar *master_hosturl,
					  const gchar *master_username)
{
	ESourceGroup *group;
	const gchar *prop;

	group = e_source_peek_group (source);
	if (!group || !e_source_group_peek_base_uri (group) ||
	    g_ascii_strcasecmp (e_source_group_peek_base_uri (group), EWS_BASE_URI) != 0)
		return FALSE;

	if (!master_hosturl && !master_username)
		return TRUE;

	prop = e_source_get_property (source, "hosturl");
	if (master_hosturl && (!prop || g_ascii_strcasecmp (master_hosturl, prop) != 0))
		return FALSE;

	prop = e_source_get_property (source, "username");
	if (master_username && (!prop || g_ascii_strcasecmp (master_username, prop) != 0))
		return FALSE;

	return TRUE;
}

static void
gather_ews_sources (ESourceList *source_list,
		    const gchar *master_hosturl,
		    const gchar *master_username,
		    GList **esources)
{
	GSList *giter, *siter;

	g_return_if_fail (source_list != NULL);
	g_return_if_fail (esources != NULL);

	for (giter = e_source_list_peek_groups (source_list); giter; giter = giter->next) {
		ESourceGroup *group = giter->data;

		if (group && e_source_group_peek_base_uri (group) &&
		    g_ascii_strcasecmp (e_source_group_peek_base_uri (group), EWS_BASE_URI) == 0) {
			for (siter = e_source_group_peek_sources (group); siter; siter = siter->next) {
				ESource *source = siter->data;

				if (source && e_ews_folder_utils_source_is_for_account (source, master_hosturl, master_username))
					*esources = g_list_prepend (*esources, g_object_ref (source));
			}
		}
	}
}

GList *
e_ews_folder_utils_get_esources (const gchar *master_hosturl,
				 const gchar *master_username,
				 GCancellable *cancellable,
				 GError **perror)
{
	ESourceList *contacts = NULL, *events = NULL, *memos = NULL, *tasks = NULL;
	GList *esources = NULL;

	e_ews_folder_utils_manage_source_lists (&contacts, &events, &memos, &tasks, FALSE);

	gather_ews_sources (contacts, master_hosturl, master_username, &esources);
	gather_ews_sources (events, master_hosturl, master_username, &esources);
	gather_ews_sources (memos, master_hosturl, master_username, &esources);
	gather_ews_sources (tasks, master_hosturl, master_username, &esources);

	return g_list_reverse (esources);
}

gboolean
e_ews_folder_utils_is_subscribed_as_esource (const GList *esources,
                                             const gchar *master_hosturl,
                                             const gchar *master_username,
                                             const gchar *folder_id)
{
	return e_ews_folder_utils_get_source_for_folder (esources, master_hosturl, master_username, folder_id) != NULL;
}

/* filters @esources thus the resulting list will contain ESource-s only for @profile;
 * free returned list with g_list_free_full (list, g_object_unref); */
GList *
e_ews_folder_utils_filter_sources_for_account (const GList *esources,
                                               const gchar *master_hosturl,
                                               const gchar *master_username)
{
	GList *found = NULL;
	const GList *iter;

	for (iter = esources; iter; iter = iter->next) {
		ESource *source = iter->data;

		if (e_ews_folder_utils_source_is_for_account (source, master_hosturl, master_username))
			found = g_list_prepend (found, g_object_ref (source));
	}

	return g_list_reverse (found);
}

/* returns (not-reffed) member of @esources, which is for @profile and @folder_id */
ESource *
e_ews_folder_utils_get_source_for_folder (const GList *esources,
                                          const gchar *master_hosturl,
                                          const gchar *master_username,
                                          const gchar *folder_id)
{
	const GList *iter;

	for (iter = esources; iter; iter = iter->next) {
		ESource *source = iter->data;

		if (e_ews_folder_utils_source_is_for_account (source, master_hosturl, master_username)) {
			const gchar *source_folder_id = e_source_get_property (source, "folder-id");

			if (source_folder_id && g_str_equal (source_folder_id, folder_id))
				return source;
		}
	}

	return NULL;
}
